package advisor

import (
	"encoding/json"
	"testing"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"gopkg.in/yaml.v3"
)

var mockConfigOverrideYAMLStr = `
template: bb.sql-review.prod # Provide the template id, then we can extend rules from the specific template.
ruleList:
  - type: statement.select.no-select-all
    level: DISABLED
  - type: table.drop-naming-convention
    level: WARNING
  - type: table.require-pk
    level: TEST
  - type: naming.table
    payload:
      format: "^table_[a-z]+(_[a-z]+)*$"
  - type: naming.column
    payload:
      maxLength: 24
  - type: column.required
    level: ERROR
    payload:
      columnList:
        - name
`

func TestConfigOverride(t *testing.T) {
	override := &SQLReviewConfigOverride{}
	err := yaml.Unmarshal([]byte(mockConfigOverrideYAMLStr), override)
	require.NoError(t, err)

	ruleList, err := MergeSQLReviewRules(override)
	require.NoError(t, err)

	for _, rule := range ruleList {
		switch rule.Type {
		case "statement.select.no-select-all":
			assert.Equal(t, SchemaRuleLevelDisabled, rule.Level)
		case "table.drop-naming-convention":
			assert.Equal(t, SchemaRuleLevelWarning, rule.Level)
		case "table.require-pk":
			assert.Equal(t, SchemaRuleLevelError, rule.Level)
		case "naming.table":
			assert.Equal(t, SchemaRuleLevelWarning, rule.Level)

			var nr NamingRulePayload
			err := json.Unmarshal([]byte(rule.Payload), &nr)
			require.NoError(t, err)

			assert.Equal(t, "^table_[a-z]+(_[a-z]+)*$", nr.Format)
			assert.Equal(t, 64, nr.MaxLength)
		case "naming.column":
			assert.Equal(t, SchemaRuleLevelWarning, rule.Level)

			var nr NamingRulePayload
			err := json.Unmarshal([]byte(rule.Payload), &nr)
			require.NoError(t, err)

			assert.Equal(t, "^[a-z]+(_[a-z]+)*$", nr.Format)
			assert.Equal(t, 24, nr.MaxLength)
		case "column.required":
			assert.Equal(t, SchemaRuleLevelError, rule.Level)

			var payload RequiredColumnRulePayload
			err := json.Unmarshal([]byte(rule.Payload), &payload)
			require.NoError(t, err)

			assert.Equal(t, 1, len(payload.ColumnList))
			assert.Equal(t, "name", payload.ColumnList[0])
		}
	}
}
